@@ -1,9 +1,10 @@
+{-# LANGUAGE CPP                 #-}
 {-# LANGUAGE NamedFieldPuns      #-}
 {-# LANGUAGE OverloadedStrings   #-}
 {-# LANGUAGE Rank2Types          #-}
 {-# LANGUAGE ScopedTypeVariables #-}
 {-# LANGUAGE TupleSections       #-}
-{-# LANGUAGE RecordWildCards #-}
+{-# LANGUAGE RecordWildCards     #-}
 
 {-|
 
@@ -21,6 +22,7 @@ module Hledger.Data.Journal (
   addTransactionModifier,
   addPeriodicTransaction,
   addTransaction,
+  journalDbg,
   journalInferMarketPricesFromTransactions,
   journalInferCommodityStyles,
   journalStyleAmounts,
@@ -29,8 +31,7 @@ module Hledger.Data.Journal (
   journalCommodityStylesWith,
   journalToCost,
   journalInferEquityFromCosts,
-  journalInferCostsFromEquity,
-  journalMarkRedundantCosts,
+  journalTagCostsAndEquityAndMaybeInferCosts,
   journalReverse,
   journalSetLastReadTime,
   journalRenumberAccountDeclarations,
@@ -94,11 +95,13 @@ module Hledger.Data.Journal (
   journalAccountTypes,
   journalAddAccountTypes,
   journalPostingsAddAccountTags,
+  defaultBaseConversionAccount,
   -- journalPrices,
-  journalConversionAccount,
+  journalBaseConversionAccount,
   journalConversionAccounts,
   -- * Misc
   canonicalStyleFrom,
+
   nulljournal,
   journalConcat,
   journalNumberTransactions,
@@ -121,7 +124,10 @@ import Control.Monad.State.Strict (StateT)
 import Data.Char (toUpper, isDigit)
 import Data.Default (Default(..))
 import Data.Foldable (toList)
-import Data.List ((\\), find, foldl', sortBy, union, intercalate)
+import Data.List ((\\), find, sortBy, union, intercalate)
+#if !MIN_VERSION_base(4,20,0)
+import Data.List (foldl')
+#endif
 import Data.List.Extra (nubSort)
 import qualified Data.Map.Strict as M
 import Data.Maybe (catMaybes, fromMaybe, mapMaybe, maybeToList)
@@ -134,7 +140,6 @@ import Data.Time.Clock.POSIX (POSIXTime)
 import Data.Tree (Tree(..), flatten)
 import Text.Printf (printf)
 import Text.Megaparsec (ParsecT)
-import Text.Megaparsec.Custom (FinalParseError)
 
 import Hledger.Utils
 import Hledger.Data.Types
@@ -149,6 +154,7 @@ import System.FilePath (takeFileName)
 import Data.Ord (comparing)
 import Hledger.Data.Dates (nulldate)
 import Data.List (sort)
+import Data.Function ((&))
 -- import Data.Function ((&))
 
 
@@ -178,21 +184,43 @@ instance Show Journal where
              (length $ jtxns j)
              (length accounts)
              (show accounts)
-             (show $ jinferredcommodities j)
+             (show $ jinferredcommoditystyles j)
              -- ++ (show $ journalTransactions l)
              where accounts = filter (/= "root") $ flatten $ journalAccountNameTree j
 
--- showJournalDebug j = unlines [
---                       show j
---                      ,show (jtxns j)
---                      ,show (jtxnmodifiers j)
---                      ,show (jperiodictxns j)
---                      ,show $ jparsetimeclockentries j
---                      ,show $ jpricedirectives j
---                      ,show $ jfinalcommentlines j
---                      ,show $ jparsestate j
---                      ,show $ map fst $ jfiles j
---                      ]
+journalDbg j@Journal{..} = chomp $ unlines $
+  ("Journal " ++ takeFileName (journalFilePath j)++":") :  --  ++ " {"
+  map (" "<>) [
+   "jparsedefaultyear: "         <> shw jparsedefaultyear
+  ,"jparsedefaultcommodity: "    <> shw jparsedefaultcommodity
+  ,"jparsedecimalmark: "         <> shw jparsedecimalmark
+  ,"jparseparentaccounts: "      <> shw jparseparentaccounts
+  ,"jparsealiases: "             <> shw jparsealiases
+  -- ,"jparsetimeclockentries: " <> shw jparsetimeclockentries
+  ,"jincludefilestack: "         <> shw jincludefilestack
+  ,"jdeclaredpayees: "           <> shw jdeclaredpayees
+  ,"jdeclaredtags: "             <> shw jdeclaredtags
+  ,"jdeclaredaccounts: "         <> shw jdeclaredaccounts
+  ,"jdeclaredaccounttags: "      <> shw jdeclaredaccounttags
+  ,"jdeclaredaccounttypes: "     <> shw jdeclaredaccounttypes
+  ,"jaccounttypes: "             <> shw jaccounttypes
+  ,"jdeclaredcommodities: "      <> shw jdeclaredcommodities
+  ,"jinferredcommoditystyles: "  <> shw jinferredcommoditystyles
+  ,"jglobalcommoditystyles: "    <> shw jglobalcommoditystyles
+  ,"jpricedirectives: "          <> shw jpricedirectives
+  ,"jinferredmarketprices: "     <> shw jinferredmarketprices
+  ,"jtxnmodifiers: "             <> shw jtxnmodifiers
+  -- ,"jperiodictxns: "          <> shw jperiodictxns
+  ,"jtxns: "                     <> shw jtxns
+  ,"jfinalcommentlines: "        <> shw jfinalcommentlines
+  ,"jfiles: "                    <> shw jfiles
+  ,"jlastreadtime: "             <> shw jlastreadtime
+  ]
+  -- ++ ["}"]
+  where
+    shw :: Show a => a -> String
+    shw = show
+    -- shw = pshow
 
 -- The semigroup instance for Journal is useful for two situations.
 --
@@ -232,12 +260,32 @@ journalConcat j1 j2 =
     ,jdeclaredpayees            = jdeclaredpayees            j1 <> jdeclaredpayees            j2
     ,jdeclaredtags              = jdeclaredtags              j1 <> jdeclaredtags              j2
     ,jdeclaredaccounts          = jdeclaredaccounts          j1 <> jdeclaredaccounts          j2
-    ,jdeclaredaccounttags       = jdeclaredaccounttags       j1 <> jdeclaredaccounttags       j2
-    ,jdeclaredaccounttypes      = jdeclaredaccounttypes      j1 <> jdeclaredaccounttypes      j2
-    ,jaccounttypes              = jaccounttypes              j1 <> jaccounttypes              j2
-    ,jglobalcommoditystyles     = jglobalcommoditystyles     j1 <> jglobalcommoditystyles     j2
-    ,jcommodities               = jcommodities               j1 <> jcommodities               j2
-    ,jinferredcommodities       = jinferredcommodities       j1 <> jinferredcommodities       j2
+    --
+    -- The next six fields are Maps, which need to be merged carefully for correct semantics,
+    -- especially the first two, which have list values. There may still be room for improvement here.
+    --
+    -- ,jdeclaredaccounttags   :: M.Map AccountName [Tag]
+    -- jdeclaredaccounttags can have multiple duplicated/conflicting values for an account's tag.
+    ,jdeclaredaccounttags       = M.unionWith (<>) (jdeclaredaccounttags j1) (jdeclaredaccounttags j2)
+    --
+    -- ,jdeclaredaccounttypes  :: M.Map AccountType [AccountName]
+    -- jdeclaredaccounttypes can have multiple duplicated/conflicting values for an account's type.
+    ,jdeclaredaccounttypes      = M.unionWith (<>) (jdeclaredaccounttypes j1) (jdeclaredaccounttypes j2)
+    --
+    -- ,jaccounttypes          :: M.Map AccountName AccountType
+    -- jaccounttypes has a single type for any given account. When it had multiple type declarations, the last/rightmost wins.
+    ,jaccounttypes              = M.unionWith (const id) (jaccounttypes j1) (jaccounttypes j2)
+    --
+    -- ,jglobalcommoditystyles :: M.Map CommoditySymbol AmountStyle
+    ,jglobalcommoditystyles     = (<>) (jglobalcommoditystyles j1) (jglobalcommoditystyles j2)
+    --
+    -- ,jdeclaredcommodities           :: M.Map CommoditySymbol Commodity
+    ,jdeclaredcommodities               = (<>) (jdeclaredcommodities j1) (jdeclaredcommodities j2)
+    --
+    -- ,jinferredcommoditystyles   :: M.Map CommoditySymbol AmountStyle
+    ,jinferredcommoditystyles       = (<>) (jinferredcommoditystyles j1) (jinferredcommoditystyles j2)
+    --
+    --
     ,jpricedirectives           = jpricedirectives           j1 <> jpricedirectives           j2
     ,jinferredmarketprices      = jinferredmarketprices      j1 <> jinferredmarketprices      j2
     ,jtxnmodifiers              = jtxnmodifiers              j1 <> jtxnmodifiers              j2
@@ -295,8 +343,8 @@ nulljournal = Journal {
   ,jdeclaredaccounttypes      = M.empty
   ,jaccounttypes              = M.empty
   ,jglobalcommoditystyles     = M.empty
-  ,jcommodities               = M.empty
-  ,jinferredcommodities       = M.empty
+  ,jdeclaredcommodities               = M.empty
+  ,jinferredcommoditystyles       = M.empty
   ,jpricedirectives           = []
   ,jinferredmarketprices      = []
   ,jtxnmodifiers              = []
@@ -356,11 +404,15 @@ showJournalAmountsDebug = show.map showMixedAmountOneLine.journalPostingAmounts
 
 -- | Sorted unique commodity symbols declared by commodity directives in this journal.
 journalCommoditiesDeclared :: Journal -> [CommoditySymbol]
-journalCommoditiesDeclared = M.keys . jcommodities
+journalCommoditiesDeclared = M.keys . jdeclaredcommodities
 
--- | Sorted unique commodity symbols declared or inferred from this journal.
+-- | Sorted unique commodity symbols mentioned in this journal.
 journalCommodities :: Journal -> S.Set CommoditySymbol
-journalCommodities j = M.keysSet (jcommodities j) <> M.keysSet (jinferredcommodities j)
+journalCommodities j =
+     M.keysSet (jdeclaredcommodities j)
+  <> M.keysSet (jinferredcommoditystyles j)
+  <> S.fromList (concatMap pdcommodities $ jpricedirectives j)
+      where pdcommodities pd = [pdcommodity pd, acommodity $ pdamount pd]
 
 -- | Unique transaction descriptions used in this journal.
 journalDescriptions :: Journal -> [Text]
@@ -529,29 +581,34 @@ journalAccountType Journal{jaccounttypes} = accountNameType jaccounttypes
 journalAddAccountTypes :: Journal -> Journal
 journalAddAccountTypes j = j{jaccounttypes = journalAccountTypes j}
 
+-- | An account type inherited from the parent account(s),
+-- and whether it was originally declared by an account directive (true) or inferred from an account name (false).
+type ParentAccountType = (AccountType, Bool)
+
 -- | Build a map of all known account types, explicitly declared
 -- or inferred from the account's parent or name.
 journalAccountTypes :: Journal -> M.Map AccountName AccountType
 journalAccountTypes j = M.fromList [(a,acctType) | (a, Just (acctType,_)) <- flatten t']
   where
     t = accountNameTreeFrom $ journalAccountNames j :: Tree AccountName
-    -- Map from the top of the account tree down to the leaves, propagating
-    -- account types downward. Keep track of whether the account is declared
-    -- (True), in which case the parent account should be preferred, or merely
-    -- inferred (False), in which case the inferred type should be preferred.
-    t' = settypes Nothing t :: Tree (AccountName, Maybe (AccountType, Bool))
+    -- Traverse downward through the account tree, applying any explicitly declared account types,
+    -- otherwise inferring account types from account names when possible, and propagating account types downward.
+    -- Declared account types (possibly inherited from parent) are preferred, inferred types are used as a fallback.
+    t' = setTypeHereAndBelow Nothing t :: Tree (AccountName, Maybe (AccountType, Bool))
       where
-        settypes :: Maybe (AccountType, Bool) -> Tree AccountName -> Tree (AccountName, Maybe (AccountType, Bool))
-        settypes mparenttype (Node a subs) = Node (a, mtype) (map (settypes mtype) subs)
+        declaredtypes       = M.keys $ jdeclaredaccounttypes j
+        declaredtypesbyname = journalDeclaredAccountTypes j & fmap (,True)
+        setTypeHereAndBelow :: Maybe ParentAccountType -> Tree AccountName -> Tree (AccountName, Maybe ParentAccountType)
+        setTypeHereAndBelow mparenttype (Node a subs) = Node (a, mnewtype) (map (setTypeHereAndBelow mnewtype) subs)
           where
-            mtype = M.lookup a declaredtypes <|> minferred
-              where 
-                declaredtypes = (,True) <$> journalDeclaredAccountTypes j
-                minferred = if maybe False snd mparenttype
-                            then mparenttype
-                            else (,False) <$> accountNameInferType a <|> mparenttype
+            mnewtype = mthisacctdeclaredtype <|> mparentacctdeclaredtype <|> mthisacctinferredtype <|> mparentacctinferredtype
+              where
+                mthisacctdeclaredtype   = M.lookup a declaredtypesbyname
+                mparentacctdeclaredtype = if       fromMaybe False $ snd <$> mparenttype then mparenttype else Nothing
+                mparentacctinferredtype = if not $ fromMaybe True  $ snd <$> mparenttype then mparenttype else Nothing
+                mthisacctinferredtype   = accountNameInferTypeExcept declaredtypes a & fmap (,False)  -- XXX not sure about this Except logic.. but for now, tests pass
 
--- | Build a map of the account types explicitly declared for each account.
+-- | Build a map from account names to explicitly declared account types.
 journalDeclaredAccountTypes :: Journal -> M.Map AccountName AccountType
 journalDeclaredAccountTypes Journal{jdeclaredaccounttypes} =
   M.fromList $ concat [map (,t) as | (t,as) <- M.toList jdeclaredaccounttypes]
@@ -563,18 +620,17 @@ journalPostingsAddAccountTags :: Journal -> Journal
 journalPostingsAddAccountTags j = journalMapPostings addtags j
   where addtags p = p `postingAddTags` (journalInheritedAccountTags j $ paccount p)
 
--- | The account to use for automatically generated conversion postings in this journal:
--- the first of the journalConversionAccounts.
-journalConversionAccount :: Journal -> AccountName
-journalConversionAccount = headDef defaultConversionAccount . journalConversionAccounts
+-- | The account name to use for conversion postings generated by --infer-equity.
+-- This is the first account declared with type V/Conversion,
+-- or otherwise the defaultBaseConversionAccount (equity:conversion).
+journalBaseConversionAccount :: Journal -> AccountName
+journalBaseConversionAccount = headDef defaultBaseConversionAccount . journalConversionAccounts
 
--- | All the accounts declared or inferred as Conversion type in this journal.
+-- | All the accounts in this journal which are declared or inferred as V/Conversion type.
+-- This does not include new account names which might be generated by --infer-equity, currently.
 journalConversionAccounts :: Journal -> [AccountName]
 journalConversionAccounts = M.keys . M.filter (==Conversion) . jaccounttypes
 
--- The fallback account to use for automatically generated conversion postings
--- if no account is declared with the Conversion type.
-defaultConversionAccount = "equity:conversion"
 
 -- Various kinds of filtering on journals. We do it differently depending
 -- on the command.
@@ -839,9 +895,9 @@ journalCommodityStyles j =
   globalstyles <> declaredstyles <> defaultcommoditystyle <> inferredstyles
   where
     globalstyles          = jglobalcommoditystyles j
-    declaredstyles        = M.mapMaybe cformat $ jcommodities j
+    declaredstyles        = M.mapMaybe cformat $ jdeclaredcommodities j
     defaultcommoditystyle = M.fromList $ catMaybes [jparsedefaultcommodity j]
-    inferredstyles        = jinferredcommodities j
+    inferredstyles        = jinferredcommoditystyles j
 
 -- | Like journalCommodityStyles, but attach a particular rounding strategy to the styles,
 -- affecting how they will affect display precisions when applied.
@@ -849,14 +905,13 @@ journalCommodityStylesWith :: Rounding -> Journal -> M.Map CommoditySymbol Amoun
 journalCommodityStylesWith r = amountStylesSetRounding r . journalCommodityStyles
 
 -- | Collect and save inferred amount styles for each commodity based on
--- the posting amounts in that commodity (excluding price amounts), ie:
--- "the format of the first amount, adjusted to the highest precision of all amounts".
+-- the posting amounts in that commodity (excluding price amounts).
 -- Can return an error message eg if inconsistent number formats are found.
 journalInferCommodityStyles :: Journal -> Either String Journal
 journalInferCommodityStyles j =
   case commodityStylesFromAmounts $ journalStyleInfluencingAmounts j of
     Left e   -> Left e
-    Right cs -> Right j{jinferredcommodities = dbg7 "journalInferCommodityStyles" cs}
+    Right cs -> Right j{jinferredcommoditystyles = dbg7 "journalInferCommodityStyles" cs}
 
 -- | Given a list of amounts, in parse order (roughly speaking; see journalStyleInfluencingAmounts),
 -- build a map from their commodity names to standard commodity
@@ -929,7 +984,7 @@ canonicalStyle a b = a{asprecision=prec, asdecimalmark=decmark, asdigitgroups=mg
 journalInferMarketPricesFromTransactions :: Journal -> Journal
 journalInferMarketPricesFromTransactions j =
   j{jinferredmarketprices =
-       dbg4 "jinferredmarketprices" .
+       dbg4With (("jinferredmarketprices:\n"<>) . showMarketPrices) $
        map priceDirectiveToMarketPrice .
        concatMap postingPriceDirectivesFromCost $
        journalPostings j
@@ -939,32 +994,23 @@ journalInferMarketPricesFromTransactions j =
 journalToCost :: ConversionOp -> Journal -> Journal
 journalToCost cost j@Journal{jtxns=ts} = j{jtxns=map (transactionToCost cost) ts}
 
+-- | Identify and tag (1) equity conversion postings and (2) postings which have (or could have ?) redundant costs.
+-- And if the addcosts flag is true, also add any costs which can be inferred from equity conversion postings.
+-- This is always called before transaction balancing to tag the redundant-cost postings so they can be ignored.
+-- With --infer-costs, it is called again after transaction balancing (when it has more information to work with) to infer costs from equity postings.
+-- See transactionTagCostsAndEquityAndMaybeInferCosts for more details, and hledger manual > Cost reporting for more background.
+journalTagCostsAndEquityAndMaybeInferCosts :: Bool -> Bool -> Journal -> Either String Journal
+journalTagCostsAndEquityAndMaybeInferCosts verbosetags addcosts j = do
+  let conversionaccts = journalConversionAccounts j
+  ts <- mapM (transactionTagCostsAndEquityAndMaybeInferCosts verbosetags addcosts conversionaccts) $ jtxns j
+  return j{jtxns=ts}
+
 -- | Add equity postings inferred from costs, where needed and possible.
 -- See hledger manual > Cost reporting.
 journalInferEquityFromCosts :: Bool -> Journal -> Journal
-journalInferEquityFromCosts verbosetags j = journalMapTransactions (transactionAddInferredEquityPostings verbosetags equityAcct) j
-  where
-    equityAcct = journalConversionAccount j
-
--- | Add costs inferred from equity conversion postings, where needed and possible.
--- See hledger manual > Cost reporting.
-journalInferCostsFromEquity :: Journal -> Either String Journal
-journalInferCostsFromEquity j = do
-  ts <- mapM (transactionInferCostsFromEquity False conversionaccts) $ jtxns j
-  return j{jtxns=ts}
-  where conversionaccts = journalConversionAccounts j
-
--- XXX duplication of the above
--- | Do just the internal tagging that is normally done by journalInferCostsFromEquity,
--- identifying equity conversion postings and, in particular, postings which have redundant costs.
--- Tagging the latter is useful as it allows them to be ignored during transaction balancedness checking.
--- And that allows journalInferCostsFromEquity to be postponed till after transaction balancing,
--- when it will have more information (amounts) to work with.
-journalMarkRedundantCosts :: Journal -> Either String Journal
-journalMarkRedundantCosts j = do
-  ts <- mapM (transactionInferCostsFromEquity True conversionaccts) $ jtxns j
-  return j{jtxns=ts}
-  where conversionaccts = journalConversionAccounts j
+journalInferEquityFromCosts verbosetags j =
+  journalMapTransactions (transactionInferEquityPostings verbosetags equityAcct) j
+  where equityAcct = journalBaseConversionAccount j
 
 -- -- | Get this journal's unique, display-preference-canonicalised commodities, by symbol.
 -- journalCanonicalCommodities :: Journal -> M.Map String CommoditySymbol
@@ -1199,7 +1245,7 @@ samplejournalMaybeExplicit explicit = nulljournal
          {jtxns = [
            txnTieKnot $ Transaction {
              tindex=0,
-             tsourcepos=nullsourcepos,
+             tsourcepos=nullsourcepospair,
              tdate=fromGregorian 2008 01 01,
              tdate2=Nothing,
              tstatus=Unmarked,
@@ -1216,7 +1262,7 @@ samplejournalMaybeExplicit explicit = nulljournal
           ,
            txnTieKnot $ Transaction {
              tindex=0,
-             tsourcepos=nullsourcepos,
+             tsourcepos=nullsourcepospair,
              tdate=fromGregorian 2008 06 01,
              tdate2=Nothing,
              tstatus=Unmarked,
@@ -1233,7 +1279,7 @@ samplejournalMaybeExplicit explicit = nulljournal
           ,
            txnTieKnot $ Transaction {
              tindex=0,
-             tsourcepos=nullsourcepos,
+             tsourcepos=nullsourcepospair,
              tdate=fromGregorian 2008 06 02,
              tdate2=Nothing,
              tstatus=Unmarked,
@@ -1250,7 +1296,7 @@ samplejournalMaybeExplicit explicit = nulljournal
           ,
            txnTieKnot $ Transaction {
              tindex=0,
-             tsourcepos=nullsourcepos,
+             tsourcepos=nullsourcepospair,
              tdate=fromGregorian 2008 06 03,
              tdate2=Nothing,
              tstatus=Cleared,
@@ -1267,7 +1313,7 @@ samplejournalMaybeExplicit explicit = nulljournal
           ,
            txnTieKnot $ Transaction {
              tindex=0,
-             tsourcepos=nullsourcepos,
+             tsourcepos=nullsourcepospair,
              tdate=fromGregorian 2008 10 01,
              tdate2=Nothing,
              tstatus=Unmarked,
@@ -1283,7 +1329,7 @@ samplejournalMaybeExplicit explicit = nulljournal
           ,
            txnTieKnot $ Transaction {
              tindex=0,
-             tsourcepos=nullsourcepos,
+             tsourcepos=nullsourcepospair,
              tdate=fromGregorian 2008 12 31,
              tdate2=Nothing,
              tstatus=Unmarked,
