@@ -33,17 +33,20 @@ module Hledger.Data.Types (
 where
 
 import GHC.Generics (Generic)
+import Data.Bifunctor (first)
 import Data.Decimal (Decimal, DecimalRaw(..))
 import Data.Default (Default(..))
 import Data.Functor (($>))
-import Data.List (intercalate)
+import Data.List (intercalate, sortBy)
 --XXX https://hackage.haskell.org/package/containers/docs/Data-Map.html
 --Note: You should use Data.Map.Strict instead of this module if:
 --You will eventually need all the values stored.
 --The stored values don't represent large virtual data structures to be lazily computed.
 import qualified Data.Map as M
 import Data.Ord (comparing)
+import Data.Semigroup (Min(..))
 import Data.Text (Text)
+import qualified Data.Text as T
 import Data.Time.Calendar (Day)
 import Data.Time.Clock.POSIX (POSIXTime)
 import Data.Time.LocalTime (LocalTime)
@@ -53,6 +56,7 @@ import Text.Megaparsec (SourcePos(SourcePos), mkPos)
 
 import Hledger.Utils.Regex
 
+
 -- synonyms for various date-related scalars
 #if MIN_VERSION_time(1,11,0)
 import Data.Time.Calendar (Year)
@@ -116,7 +120,7 @@ data DateSpan = DateSpan (Maybe EFDay) (Maybe EFDay) deriving (Eq,Ord,Generic)
 
 instance Default DateSpan where def = DateSpan Nothing Nothing
 
--- Typical report periods (spans of time), both finite and open-ended.
+-- Some common report subperiods, both finite and open-ended.
 -- A higher-level abstraction than DateSpan.
 data Period =
     DayPeriod Day
@@ -132,16 +136,8 @@ data Period =
 
 instance Default Period where def = PeriodAll
 
----- Typical report period/subperiod durations, from a day to a year.
---data Duration =
---    DayLong
---   WeekLong
---   MonthLong
---   QuarterLong
---   YearLong
---  deriving (Eq,Ord,Show,Generic)
-
--- Ways in which a period can be divided into subperiods.
+-- All the kinds of report interval allowed in a period expression
+-- (to generate periodic reports or periodic transactions).
 data Interval =
     NoInterval
   | Days Int
@@ -149,13 +145,10 @@ data Interval =
   | Months Int
   | Quarters Int
   | Years Int
-  | DayOfMonth Int
-  | WeekdayOfMonth Int Int
-  | DaysOfWeek [Int]
-  | DayOfYear Int Int -- Month, Day
-  -- WeekOfYear Int
-  -- MonthOfYear Int
-  -- QuarterOfYear Int
+  | NthWeekdayOfMonth Int Int  -- n,              weekday 1-7
+  | MonthDay Int               -- 1-31
+  | MonthAndDay Int Int        -- month 1-12,     monthday 1-31
+  | DaysOfWeek [Int]           -- [weekday 1-7]
   deriving (Eq,Show,Ord,Generic)
 
 instance Default Interval where def = NoInterval
@@ -164,6 +157,19 @@ type Payee = Text
 
 type AccountName = Text
 
+-- A specification indicating how to depth-limit
+data DepthSpec = DepthSpec {
+  dsFlatDepth    :: Maybe Int,
+  dsRegexpDepths :: [(Regexp, Int)]
+  } deriving (Eq,Show)
+
+-- Semigroup instance consider all regular expressions, but take the minimum of the simple flat depths
+instance Semigroup DepthSpec where
+    DepthSpec d1 l1 <> DepthSpec d2 l2 = DepthSpec (getMin <$> (Min <$> d1) <> (Min <$> d2)) (l1 ++ l2)
+
+instance Monoid DepthSpec where
+    mempty = DepthSpec Nothing []
+
 data AccountType =
     Asset
   | Liability
@@ -400,8 +406,32 @@ data PostingType = RegularPosting | VirtualPosting | BalancedVirtualPosting
 type TagName = Text
 type TagValue = Text
 type Tag = (TagName, TagValue)  -- ^ A tag name and (possibly empty) value.
+type HiddenTag = Tag            -- ^ A tag whose name begins with _.
 type DateTag = (TagName, Day)
 
+-- | Add the _ prefix to a normal visible tag's name, making it a hidden tag.
+toHiddenTag :: Tag -> HiddenTag
+toHiddenTag = first toHiddenTagName
+
+-- | Drop the _ prefix from a hidden tag's name, making it a normal visible tag.
+toVisibleTag :: HiddenTag -> Tag
+toVisibleTag = first toVisibleTagName
+
+-- | Does this tag name begin with the hidden tag prefix (_) ?
+isHiddenTagName :: TagName -> Bool
+isHiddenTagName t =
+  case T.uncons t of
+    Just ('_',_) -> True
+    _ -> False
+
+-- | Add the _ prefix to a normal visible tag's name, making it a hidden tag.
+toHiddenTagName :: TagName -> TagName
+toHiddenTagName = T.cons '_'
+
+-- | Drop the _ prefix from a hidden tag's name, making it a normal visible tag.
+toVisibleTagName :: TagName -> TagName
+toVisibleTagName = T.drop 1
+
 -- | The status of a transaction or posting, recorded with a status mark
 -- (nothing, !, or *). What these mean is ultimately user defined.
 data Status = Unmarked | Pending | Cleared
@@ -412,6 +442,12 @@ instance Show Status where -- custom show.. bad idea.. don't do it..
   show Pending   = "!"
   show Cleared   = "*"
 
+nullsourcepos :: SourcePos
+nullsourcepos = SourcePos "" (mkPos 1) (mkPos 1)
+
+nullsourcepospair :: (SourcePos, SourcePos)
+nullsourcepospair = (SourcePos "" (mkPos 1) (mkPos 1), SourcePos "" (mkPos 2) (mkPos 1))
+
 -- | A balance assertion is a declaration about an account's expected balance
 -- at a certain point (posting date and parse order). They provide additional
 -- error checking and readability to a journal file.
@@ -556,7 +592,8 @@ data TimeclockEntry = TimeclockEntry {
 -- It declares two things: a historical exchange rate between two commodities,
 -- and an amount display style for the second commodity.
 data PriceDirective = PriceDirective {
-   pddate      :: Day
+   pdsourcepos :: SourcePos
+  ,pddate      :: Day
   ,pdcommodity :: CommoditySymbol
   ,pdamount    :: Amount
   } deriving (Eq,Ord,Generic,Show)
@@ -570,50 +607,52 @@ data MarketPrice = MarketPrice {
   ,mprate :: Quantity           -- ^ One unit of the "from" commodity is worth this quantity of the "to" commodity.
   } deriving (Eq,Ord,Generic, Show)
 
+showMarketPrice MarketPrice{..} = unwords [show mpdate, T.unpack mpfrom <> ">" <> T.unpack mpto, show mprate]
+showMarketPrices = intercalate "\n" . map ((' ':).showMarketPrice) . sortBy (comparing mpdate)
+
 -- additional valuation-related types in Valuation.hs
 
--- | A Journal, containing transactions and various other things.
--- The basic data model for hledger.
+-- | A journal, containing general ledger transactions; also directives and various other things.
+-- This is hledger's main data model.
 --
--- This is used during parsing (as the type alias ParsedJournal), and
--- then finalised/validated for use as a Journal. Some extra
--- parsing-related fields are included for convenience, at least for
--- now. In a ParsedJournal these are updated as parsing proceeds, in a
--- Journal they represent the final state at end of parsing (used eg
--- by the add command).
+-- During parsing, it is used as the type alias "ParsedJournal".
+-- The jparse* fields are mainly used during parsing and included here for convenience.
+-- The list fields described as "in parse order" are usually reversed for efficiency during parsing.
+-- After parsing, "journalFinalise" converts ParsedJournal to a finalised "Journal",
+-- which has all lists correctly ordered, and much data inference and validation applied.
 --
 data Journal = Journal {
-  -- parsing-related data
-   jparsedefaultyear      :: Maybe Year                            -- ^ the current default year, specified by the most recent Y directive (or current date)
-  ,jparsedefaultcommodity :: Maybe (CommoditySymbol,AmountStyle)   -- ^ the current default commodity and its format, specified by the most recent D directive
-  ,jparsedecimalmark      :: Maybe DecimalMark                     -- ^ the character to always parse as decimal point, if set by CsvReader's decimal-mark (or a future journal directive)
-  ,jparseparentaccounts   :: [AccountName]                         -- ^ the current stack of parent account names, specified by apply account directives
-  ,jparsealiases          :: [AccountAlias]                        -- ^ the current account name aliases in effect, specified by alias directives (& options ?)
+  -- parsing-related state
+   jparsedefaultyear        :: Maybe Year                             -- ^ the current default year, specified by the most recent Y directive (or current date)
+  ,jparsedefaultcommodity   :: Maybe (CommoditySymbol,AmountStyle)    -- ^ the current default commodity and its format, specified by the most recent D directive
+  ,jparsedecimalmark        :: Maybe DecimalMark                      -- ^ the character to always parse as decimal point, if set by CsvReader's decimal-mark (or a future journal directive)
+  ,jparseparentaccounts     :: [AccountName]                          -- ^ the current stack of parent account names, specified by apply account directives
+  ,jparsealiases            :: [AccountAlias]                         -- ^ the current account name aliases in effect, specified by alias directives (& options ?)
   -- ,jparsetransactioncount :: Integer                               -- ^ the current count of transactions parsed so far (only journal format txns, currently)
-  ,jparsetimeclockentries :: [TimeclockEntry]                       -- ^ timeclock sessions which have not been clocked out
-  ,jincludefilestack      :: [FilePath]
+  ,jparsetimeclockentries   :: [TimeclockEntry]                       -- ^ timeclock sessions which have not been clocked out
+  ,jincludefilestack        :: [FilePath]
   -- principal data
-  ,jdeclaredpayees        :: [(Payee,PayeeDeclarationInfo)]         -- ^ Payees declared by payee directives, in parse order (after journal finalisation)
-  ,jdeclaredtags          :: [(TagName,TagDeclarationInfo)]         -- ^ Tags declared by tag directives, in parse order (after journal finalisation)
-  ,jdeclaredaccounts      :: [(AccountName,AccountDeclarationInfo)] -- ^ Accounts declared by account directives, in parse order (after journal finalisation)
-  ,jdeclaredaccounttags   :: M.Map AccountName [Tag]                -- ^ Accounts which have tags declared in their directives, and those tags. (Does not include parents' tags.)
-  ,jdeclaredaccounttypes  :: M.Map AccountType [AccountName]        -- ^ Accounts whose type has been explicitly declared in their account directives, grouped by type.
-  ,jaccounttypes          :: M.Map AccountName AccountType          -- ^ All accounts for which a type has been declared or can be inferred from its parent or its name.
-  ,jglobalcommoditystyles :: M.Map CommoditySymbol AmountStyle      -- ^ per-commodity display styles declared globally, eg by command line option or import command
-  ,jcommodities           :: M.Map CommoditySymbol Commodity        -- ^ commodities and formats declared by commodity directives
-  ,jinferredcommodities   :: M.Map CommoditySymbol AmountStyle      -- ^ commodities and formats inferred from journal amounts
-  ,jpricedirectives       :: [PriceDirective]                       -- ^ Declarations of market prices by P directives, in parse order (after journal finalisation)
-  ,jinferredmarketprices  :: [MarketPrice]                          -- ^ Market prices implied by transactions, in parse order (after journal finalisation)
-  ,jtxnmodifiers          :: [TransactionModifier]
-  ,jperiodictxns          :: [PeriodicTransaction]
-  ,jtxns                  :: [Transaction]
-  ,jfinalcommentlines     :: Text                                   -- ^ any final trailing comments in the (main) journal file
-  ,jfiles                 :: [(FilePath, Text)]                     -- ^ the file path and raw text of the main and
-                                                                    --   any included journal files. The main file is first,
-                                                                    --   followed by any included files in the order encountered.
-                                                                    --   TODO: FilePath is a sloppy type here, don't assume it's a
-                                                                    --   real file; values like "", "-", "(string)" can be seen
-  ,jlastreadtime          :: POSIXTime                              -- ^ when this journal was last read from its file(s)
+  ,jdeclaredpayees          :: [(Payee,PayeeDeclarationInfo)]         -- ^ Payees declared by payee directives, in parse order.
+  ,jdeclaredtags            :: [(TagName,TagDeclarationInfo)]         -- ^ Tags declared by tag directives, in parse order.
+  ,jdeclaredaccounts        :: [(AccountName,AccountDeclarationInfo)] -- ^ Accounts declared by account directives, in parse order.
+  ,jdeclaredaccounttags     :: M.Map AccountName [Tag]                -- ^ Accounts which were declared with tags, and those tags.
+  ,jdeclaredaccounttypes    :: M.Map AccountType [AccountName]        -- ^ Accounts which were declared with a type: tag, grouped by the type.
+  ,jaccounttypes            :: M.Map AccountName AccountType          -- ^ All the account types known, from account declarations or account names or parent accounts.
+  ,jdeclaredcommodities     :: M.Map CommoditySymbol Commodity        -- ^ Commodities (and their display styles) declared by commodity directives, in parse order.
+  ,jinferredcommoditystyles :: M.Map CommoditySymbol AmountStyle      -- ^ Commodity display styles inferred from amounts in the journal.
+  ,jglobalcommoditystyles   :: M.Map CommoditySymbol AmountStyle      -- ^ Commodity display styles declared by command line options (sometimes augmented, see the import command).
+  ,jpricedirectives         :: [PriceDirective]                       -- ^ P (market price) directives in the journal, in parse order.
+  ,jinferredmarketprices    :: [MarketPrice]                          -- ^ Market prices inferred from transactions in the journal, in parse order.
+  ,jtxnmodifiers            :: [TransactionModifier]                  -- ^ Auto posting rules declared in the journal.
+  ,jperiodictxns            :: [PeriodicTransaction]                  -- ^ Periodic transaction rules declared in the journal.
+  ,jtxns                    :: [Transaction]                          -- ^ Transactions recorded in the journal. The important bit.
+  ,jfinalcommentlines       :: Text                                   -- ^ any final trailing comments in the (main) journal file
+  ,jfiles                   :: [(FilePath, Text)]                     -- ^ the file path and raw text of the main and
+                                                                      --   any included journal files. The main file is first,
+                                                                      --   followed by any included files in the order encountered.
+                                                                      --   TODO: FilePath is a sloppy type here, don't assume it's a
+                                                                      --   real file; values like "", "-", "(string)" can be seen
+  ,jlastreadtime            :: POSIXTime                              -- ^ when this journal was last read from its file(s)
   -- NOTE: after adding new fields, eg involving account names, consider updating
   -- the Anon instance in Hleger.Cli.Anon
   } deriving (Eq, Generic)
