@@ -107,7 +107,6 @@ module Hledger.Read (
   orDieTrying,
 
   -- * Misc
-  journalStrictChecks,
   saveLatestDates,
   saveLatestDatesForFiles,
 
@@ -126,7 +125,7 @@ module Hledger.Read (
 
 --- ** imports
 import qualified Control.Exception as C
-import Control.Monad (unless, when)
+import Control.Monad (unless, when, forM)
 import "mtl" Control.Monad.Except (ExceptT(..), runExceptT, liftEither)
 import Control.Monad.IO.Class (MonadIO, liftIO)
 import Data.Default (def)
@@ -159,7 +158,8 @@ import Hledger.Read.RulesReader (tests_RulesReader)
 -- import Hledger.Read.TimeclockReader (tests_TimeclockReader)
 import Hledger.Utils
 import Prelude hiding (getContents, writeFile)
-import Hledger.Data.JournalChecks (journalCheckAccounts, journalCheckCommodities)
+import Hledger.Data.JournalChecks (journalStrictChecks)
+import Text.Printf (printf)
 
 --- ** doctest setup
 -- $setup
@@ -220,15 +220,16 @@ type PrefixedFilePath = FilePath
 -- we use the journal reader (for predictability).
 --
 readJournal :: InputOpts -> Maybe FilePath -> Text -> ExceptT String IO Journal
-readJournal iopts@InputOpts{strict_} mpath txt = do
+readJournal iopts@InputOpts{strict_, _defer} mpath txt = do
   let r :: Reader IO = fromMaybe JournalReader.reader $ findReader (mformat_ iopts) mpath
   dbg6IO "readJournal: trying reader" (rFormat r)
   j <- rReadFn r iopts (fromMaybe "(string)" mpath) txt
-  when strict_ $ liftEither $ journalStrictChecks j
+  when (strict_ && not _defer) $ liftEither $ journalStrictChecks j
   return j
 
 -- | Read a Journal from this file, or from stdin if the file path is -,
 -- with strict checks if enabled, or return an error message.
+-- (Note strict checks are disabled temporarily here when this is called by readJournalFiles).
 -- The file path can have a READER: prefix.
 --
 -- The reader (data format) to use is determined from (in priority order):
@@ -241,15 +242,14 @@ readJournal iopts@InputOpts{strict_} mpath txt = do
 -- generation, a rules file for converting CSV data, etc.
 --
 -- If using --new, and if latest-file writing is enabled in input options,
--- and after passing strict checks if enabled, a .latest.FILE file will be created/updated
--- (for the main file only, not for included files),
--- to remember the latest transaction date (and how many transactions on this date)
--- successfully read.
+-- and not deferred by readJournalFiles, and after passing strict checks if enabled,
+-- a .latest.FILE file will be created/updated (for the main file only, not for included files),
+-- to remember the latest transaction date processed.
 --
 readJournalFile :: InputOpts -> PrefixedFilePath -> ExceptT String IO Journal
-readJournalFile iopts@InputOpts{new_, new_save_} prefixedfile = do
+readJournalFile iopts@InputOpts{new_, new_save_, _defer} prefixedfile = do
   (j, mlatestdates) <- readJournalFileAndLatestDates iopts prefixedfile
-  when (new_ && new_save_) $ liftIO $
+  when (new_ && new_save_ && not _defer) $ liftIO $
     case mlatestdates of
       Nothing                        -> return ()
       Just (LatestDatesForFile f ds) -> saveLatestDates ds f
@@ -279,9 +279,6 @@ readJournalFileAndLatestDates iopts prefixedfile = do
 
 -- | Read a Journal from each specified file path (using @readJournalFile@) 
 -- and combine them into one; or return the first error message.
--- Strict checks, if enabled, are deferred till the end.
--- Writing .latest files, if enabled, is also deferred till the end,
--- and happens only if strict checks pass.
 --
 -- Combining Journals means concatenating them, basically.
 -- The parse state resets at the start of each file, which means that
@@ -289,9 +286,16 @@ readJournalFileAndLatestDates iopts prefixedfile = do
 -- They do affect included child files though.
 -- Also the final parse state saved in the Journal does span all files.
 --
+-- Strict checks, if enabled, are temporarily deferred until all files are read,
+-- to ensure they see the whole journal, and/or to avoid redundant work.
+-- (Some checks, like assertions and ordereddates, might still be doing redundant work ?)
+--
+-- Writing .latest files, if enabled, is also deferred till the end,
+-- and is done only if strict checks pass.
+--
 readJournalFiles :: InputOpts -> [PrefixedFilePath] -> ExceptT String IO Journal
-readJournalFiles iopts@InputOpts{strict_,new_,new_save_} prefixedfiles = do
-  let iopts' = iopts{strict_=False, new_save_=False}
+readJournalFiles iopts@InputOpts{strict_, new_, new_save_} prefixedfiles = do
+  let iopts' = iopts{_defer=True}
   (j, latestdatesforfiles) <-
     traceOrLogAt 6 ("readJournalFiles: "++show prefixedfiles) $
     readJournalFilesAndLatestDates iopts' prefixedfiles
@@ -307,13 +311,6 @@ readJournalFilesAndLatestDates iopts pfs = do
   (js, lastdates) <- unzip <$> mapM (readJournalFileAndLatestDates iopts) pfs
   return (maybe def sconcat $ nonEmpty js, catMaybes lastdates)
 
--- | Run the extra -s/--strict checks on a journal,
--- returning the first error message if any of them fail.
-journalStrictChecks :: Journal -> Either String ()
-journalStrictChecks j = do
-  journalCheckAccounts j
-  journalCheckCommodities j
-
 -- | An easy version of 'readJournal' which assumes default options, and fails
 -- in the IO monad.
 readJournal' :: Text -> IO Journal
@@ -405,12 +402,15 @@ saveLatestDatesForFiles = mapM_ (\(LatestDatesForFile f ds) -> saveLatestDates d
 previousLatestDates :: FilePath -> IO LatestDates
 previousLatestDates f = do
   let latestfile = latestDatesFileFor f
-      parsedate s = maybe (fail $ "could not parse date \"" ++ s ++ "\"") return $
-                      parsedateM s
   exists <- doesFileExist latestfile
-  if exists
-  then traverse (parsedate . T.unpack . T.strip) . T.lines =<< readFileStrictly latestfile
-  else return []
+  t <- if exists then readFileStrictly latestfile else return T.empty
+  let nls = zip [1::Int ..] $ T.lines t
+  fmap catMaybes $ forM nls $ \(n,l) -> do
+    let s = T.unpack $ T.strip l
+    case (s, parsedateM s) of
+      ("", _)       -> return Nothing
+      (_,  Nothing) -> error' (printf "%s:%d: invalid date: \"%s\"" latestfile n s)
+      (_,  Just d)  -> return $ Just d
 
 -- | Where to save latest transaction dates for the given file path.
 -- (.latest.FILE)
