@@ -91,7 +91,6 @@ import Data.Time.LocalTime
 import Safe
 import Text.Megaparsec hiding (parse)
 import Text.Megaparsec.Char
-import Text.Megaparsec.Custom
 import Text.Printf
 import System.FilePath
 import "Glob" System.FilePath.Glob hiding (match)
@@ -291,7 +290,7 @@ includedirectivep :: MonadIO m => ErroringJournalParser m ()
 includedirectivep = do
   string "include"
   lift skipNonNewlineSpaces1
-  prefixedglob <- T.unpack <$> takeWhileP Nothing (/= '\n') -- don't consume newline yet
+  prefixedglob <- rstrip . T.unpack <$> takeWhileP Nothing (/= '\n') -- don't consume newline yet
   parentoff <- getOffset
   parentpos <- getSourcePos
   let (mprefix,glb) = splitReaderPrefix prefixedglob
@@ -375,7 +374,7 @@ includedirectivep = do
       ,jparseparentaccounts   = jparseparentaccounts j
       ,jparsedecimalmark      = jparsedecimalmark j
       ,jparsealiases          = jparsealiases j
-      ,jcommodities           = jcommodities j
+      ,jdeclaredcommodities           = jdeclaredcommodities j
       -- ,jparsetransactioncount = jparsetransactioncount j
       ,jparsetimeclockentries = jparsetimeclockentries j
       ,jincludefilestack      = filepath : jincludefilestack j
@@ -510,7 +509,7 @@ commoditydirectiveonelinep = do
   let comm = Commodity{csymbol=acommodity, cformat=Just $ dbg6 "style from commodity directive" astyle}
   if isNothing $ asdecimalmark astyle
   then customFailure $ parseErrorAt off pleaseincludedecimalpoint
-  else modify' (\j -> j{jcommodities=M.insert acommodity comm $ jcommodities j})
+  else modify' (\j -> j{jdeclaredcommodities=M.insert acommodity comm $ jdeclaredcommodities j})
 
 pleaseincludedecimalpoint :: String
 pleaseincludedecimalpoint = chomp $ unlines [
@@ -536,7 +535,7 @@ commoditydirectivemultilinep = do
   subdirectives <- many $ indented (eitherP (formatdirectivep sym) (lift restofline))
   let mfmt = lastMay $ lefts subdirectives
   let comm = Commodity{csymbol=sym, cformat=mfmt}
-  modify' (\j -> j{jcommodities=M.insert sym comm $ jcommodities j})
+  modify' (\j -> j{jdeclaredcommodities=M.insert sym comm $ jdeclaredcommodities j})
   where
     indented = (lift skipNonNewlineSpaces1 >>)
 
@@ -672,15 +671,16 @@ defaultcommoditydirectivep = do
 
 marketpricedirectivep :: JournalParser m PriceDirective
 marketpricedirectivep = do
+  pos <- getSourcePos
   char 'P' <?> "market price"
   lift skipNonNewlineSpaces
   date <- try (do {LocalTime d _ <- datetimep; return d}) <|> datep -- a time is ignored
   lift skipNonNewlineSpaces1
   symbol <- lift commoditysymbolp
-  lift skipNonNewlineSpaces
+  lift skipNonNewlineSpaces1
   price <- amountp
   lift restofline
-  return $ PriceDirective date symbol price
+  return $ PriceDirective pos date symbol price
 
 ignoredpricecommoditydirectivep :: JournalParser m ()
 ignoredpricecommoditydirectivep = do
@@ -1117,7 +1117,7 @@ tests_JournalReader = testGroup "JournalReader" [
         [("a:b", AccountDeclarationInfo{adicomment          = "type:asset\n"
                                        ,aditags             = [("type","asset")]
                                        ,adideclarationorder = 1
-                                       ,adisourcepos        = fst nullsourcepos
+                                       ,adisourcepos        = nullsourcepos
                                        })
         ]
       ]
@@ -1146,6 +1146,7 @@ tests_JournalReader = testGroup "JournalReader" [
   ,testCase "marketpricedirectivep" $ assertParseEq marketpricedirectivep
     "P 2017/01/30 BTC $922.83\n"
     PriceDirective{
+      pdsourcepos = nullsourcepos,
       pddate      = fromGregorian 2017 1 30,
       pdcommodity = "BTC",
       pdamount    = usd 922.83
