@@ -21,6 +21,9 @@ module Hledger.Reports.ReportOptions (
   HasReportOpts(..),
   ReportSpec(..),
   HasReportSpec(..),
+  SortField(..),
+  SortSpec,
+  sortKeysDescription,
   overEither,
   setEither,
   BalanceCalculation(..),
@@ -31,6 +34,7 @@ module Hledger.Reports.ReportOptions (
   defreportopts,
   rawOptsToReportOpts,
   defreportspec,
+  defsortspec,
   setDefaultConversionOp,
   reportOptsToSpec,
   updateReportSpec,
@@ -71,15 +75,14 @@ import Data.Char (toLower)
 import Data.Either (fromRight)
 import Data.Either.Extra (eitherToMaybe)
 import Data.Functor.Identity (Identity(..))
-import Data.List.Extra (find, isPrefixOf, nubSort)
+import Data.List (partition)
+import Data.List.Extra (find, isPrefixOf, nubSort, stripPrefix)
 import Data.Maybe (fromMaybe, isJust, isNothing)
 import qualified Data.Text as T
 import Data.Time.Calendar (Day, addDays)
 import Data.Default (Default(..))
 import Safe (headMay, lastDef, lastMay, maximumMay, readMay)
 
-import Text.Megaparsec.Custom
-
 import Hledger.Data
 import Hledger.Query
 import Hledger.Utils
@@ -132,18 +135,21 @@ data ReportOpts = ReportOpts {
     ,conversionop_     :: Maybe ConversionOp  -- ^ Which operation should we apply to conversion transactions?
     ,value_            :: Maybe ValuationType  -- ^ What value should amounts be converted to ?
     ,infer_prices_     :: Bool      -- ^ Infer market prices from transactions ?
-    ,depth_            :: Maybe Int
+    ,depth_            :: DepthSpec
     ,date2_            :: Bool
     ,empty_            :: Bool
     ,no_elide_         :: Bool
     ,real_             :: Bool
     ,format_           :: StringFormat
+    ,balance_base_url_ :: Maybe T.Text
     ,pretty_           :: Bool
     ,querystring_      :: [T.Text]
     --
     ,average_          :: Bool
     -- for posting reports (register)
     ,related_          :: Bool
+    -- for sorting reports (register)
+    ,sortspec_             :: SortSpec
     -- for account transactions reports (aregister)
     ,txn_dates_        :: Bool
     -- for balance reports (bal, bs, cf, is)
@@ -189,16 +195,18 @@ defreportopts = ReportOpts
     , conversionop_     = Nothing
     , value_            = Nothing
     , infer_prices_     = False
-    , depth_            = Nothing
+    , depth_            = DepthSpec Nothing []
     , date2_            = False
     , empty_            = False
     , no_elide_         = False
     , real_             = False
     , format_           = def
+    , balance_base_url_ = Nothing
     , pretty_           = False
     , querystring_      = []
     , average_          = False
     , related_          = False
+    , sortspec_         = defsortspec 
     , txn_dates_        = False
     , balancecalc_      = def
     , balanceaccum_     = def
@@ -219,18 +227,18 @@ defreportopts = ReportOpts
     , layout_           = LayoutWide Nothing
     }
 
--- | Generate a ReportOpts from raw command-line input, given a day.
+-- | Generate a ReportOpts from raw command-line input, given a day and whether to use ANSI colour/styles in standard output.
 -- This will fail with a usage error if it is passed
 -- - an invalid --format argument,
 -- - an invalid --value argument,
 -- - if --valuechange is called with a valuation type other than -V/--value=end.
 -- - an invalid --pretty argument,
-rawOptsToReportOpts :: Day -> RawOpts -> ReportOpts
-rawOptsToReportOpts d rawopts =
+rawOptsToReportOpts :: Day -> Bool -> RawOpts -> ReportOpts
+rawOptsToReportOpts d usecoloronstdout rawopts =
 
     let formatstring = T.pack <$> maybestringopt "format" rawopts
         querystring  = map T.pack $ listofstringopt "args" rawopts  -- doesn't handle an arg like "" right
-        pretty = fromMaybe False $ alwaysneveropt "pretty" rawopts
+        pretty = fromMaybe False $ ynopt "pretty" rawopts
 
         format = case parseStringFormat <$> formatstring of
             Nothing         -> defaultBalanceLineFormat
@@ -244,15 +252,17 @@ rawOptsToReportOpts d rawopts =
           ,conversionop_     = conversionOpFromRawOpts rawopts
           ,value_            = valuationTypeFromRawOpts rawopts
           ,infer_prices_     = boolopt "infer-market-prices" rawopts
-          ,depth_            = maybeposintopt "depth" rawopts
+          ,depth_            = depthFromRawOpts rawopts
           ,date2_            = boolopt "date2" rawopts
           ,empty_            = boolopt "empty" rawopts
           ,no_elide_         = boolopt "no-elide" rawopts
           ,real_             = boolopt "real" rawopts
           ,format_           = format
+          ,balance_base_url_ = T.pack <$> maybestringopt "base-url" rawopts
           ,querystring_      = querystring
           ,average_          = boolopt "average" rawopts
           ,related_          = boolopt "related" rawopts
+          ,sortspec_         = getSortSpec rawopts
           ,txn_dates_        = boolopt "txn-dates" rawopts
           ,balancecalc_      = balancecalcopt rawopts
           ,balanceaccum_     = balanceaccumopt rawopts
@@ -268,7 +278,7 @@ rawOptsToReportOpts d rawopts =
           ,percent_          = boolopt "percent" rawopts
           ,invert_           = boolopt "invert" rawopts
           ,pretty_           = pretty
-          ,color_            = useColorOnStdout -- a lower-level helper
+          ,color_            = usecoloronstdout
           ,transpose_        = boolopt "transpose" rawopts
           ,layout_           = layoutopt rawopts
           }
@@ -329,15 +339,15 @@ balancecalcopt =
 balanceaccumopt :: RawOpts -> BalanceAccumulation
 balanceaccumopt = fromMaybe PerPeriod . balanceAccumulationOverride
 
-alwaysneveropt :: String -> RawOpts -> Maybe Bool
-alwaysneveropt opt rawopts = case maybestringopt opt rawopts of
+ynopt :: String -> RawOpts -> Maybe Bool
+ynopt opt rawopts = case maybestringopt opt rawopts of
     Just "always" -> Just True
     Just "yes"    -> Just True
     Just "y"      -> Just True
     Just "never"  -> Just False
     Just "no"     -> Just False
     Just "n"      -> Just False
-    Just _        -> usageError "--pretty's argument should be \"yes\" or \"no\" (or y, n, always, never)"
+    Just _        -> usageError "this argument should be one of y, yes, n, no"
     _             -> Nothing
 
 balanceAccumulationOverride :: RawOpts -> Maybe BalanceAccumulation
@@ -532,6 +542,21 @@ conversionOpFromRawOpts rawopts
       | n == "value", takeWhile (/=',') v `elem` ["cost", "c"] = Just ToCost  -- keep supporting --value=cost for now
       | otherwise                                   = Nothing
 
+-- | Parse the depth arguments. This can be either a flat depth that applies to
+-- all accounts, or a regular expression and depth, which only matches certain
+-- accounts. If an account name is matched by a regular expression, then the
+-- smallest depth is used. Otherwise, if no regular expressions match, then the
+-- flat depth is used. If more than one flat depth is supplied, use only the
+-- last one.
+depthFromRawOpts :: RawOpts -> DepthSpec
+depthFromRawOpts rawopts = lastDef mempty flats <> mconcat regexps
+  where
+    (flats, regexps) = partition (\(DepthSpec f rs) -> isJust f && null rs) depthSpecs
+    depthSpecs = case mapM (parseDepthSpec . T.pack) depths of
+      Right d -> d
+      Left err -> usageError $ "Unable to parse depth specification: " ++ err
+    depths = listofstringopt "depth" rawopts
+
 -- | Select the Transaction date accessor based on --date2.
 transactionDateFn :: ReportOpts -> (Transaction -> Day)
 transactionDateFn ReportOpts{..} = if date2_ then transactionDate2 else tdate
@@ -658,13 +683,56 @@ queryFromFlags :: ReportOpts -> Query
 queryFromFlags ReportOpts{..} = simplifyQuery $ And flagsq
   where
     flagsq = consIf   Real  real_
-           . consJust Depth depth_
-           $   [ (if date2_ then Date2 else Date) $ periodAsDateSpan period_
+           . consJust Depth flatDepth
+           $ map (uncurry DepthAcct) regexpDepths
+           ++  [ (if date2_ then Date2 else Date) $ periodAsDateSpan period_
                , Or $ map StatusQ statuses_
                ]
     consIf f b = if b then (f True:) else id
+    DepthSpec flatDepth regexpDepths = depth_
     consJust f = maybe id ((:) . f)
 
+-- Methods/types needed for --sort argument
+
+-- Possible arguments taken by the --sort command
+-- Each of these takes a bool, which shows if it has been inverted
+-- (True -> has been inverted, reverse the order)
+data SortField
+    = AbsAmount' Bool
+    | Account' Bool
+    | Amount' Bool
+    | Date' Bool
+    | Description' Bool
+    deriving (Show, Eq)
+type SortSpec = [SortField]
+
+-- By default, sort by date in ascending order
+defsortspec :: SortSpec
+defsortspec = [Date' False]
+
+-- Load a SortSpec from the argument given to --sort
+-- If there is no spec given, then sort by [Date' False] by default
+getSortSpec :: RawOpts -> SortSpec
+getSortSpec opts = 
+    let opt = maybestringopt "sort" opts
+        optParser s = 
+          let terms = map strip $ splitAtElement ',' s 
+              termParser t = case trimmed of
+                "date"        -> Date'        isNegated
+                "desc"        -> Description' isNegated
+                "description" -> Description' isNegated
+                "account"     -> Account'     isNegated
+                "amount"      -> Amount'      isNegated
+                "absamount"   -> AbsAmount'   isNegated
+                _ -> error' $ "unknown --sort key " ++ t ++ ". Supported keys are: " <> sortKeysDescription <> "."
+                where isNegated = isPrefixOf "-" t
+                      trimmed = fromMaybe t (stripPrefix "-" t)
+          in map termParser terms
+    in maybe defsortspec optParser opt 
+
+-- for option's help and parse error message
+sortKeysDescription = "date, desc, account, amount, absamount"  -- 'description' is also accepted
+
 -- Report dates.
 
 -- | The effective report span is the start and end dates specified by
@@ -766,7 +834,7 @@ reportPeriodOrJournalLastDay rspec j = reportPeriodLastDay rspec <|> journalOrPr
 reportPeriodName :: BalanceAccumulation -> [DateSpan] -> DateSpan -> T.Text
 reportPeriodName balanceaccumulation spans =
   case balanceaccumulation of
-    PerPeriod -> if multiyear then showDateSpan else showDateSpanMonthAbbrev
+    PerPeriod -> if multiyear then showDateSpan else showDateSpanAbbrev
       where
         multiyear = (>1) $ length $ nubSort $ map spanStartYear spans
     _ -> maybe "" (showDate . prevday) . spanEnd
@@ -819,7 +887,7 @@ makeHledgerClassyLenses ''ReportSpec
 -- >>> _rsQuery <$> setEither querystring ["assets"] defreportspec
 -- Right (Acct (RegexpCI "assets"))
 -- >>> _rsQuery <$> setEither querystring ["(assets"] defreportspec
--- Left "This regular expression is malformed, please correct it:\n(assets"
+-- Left "This regular expression is invalid or unsupported, please correct it:\n(assets"
 -- >>> _rsQuery $ set querystring ["assets"] defreportspec
 -- Acct (RegexpCI "assets")
 -- >>> _rsQuery $ set querystring ["(assets"] defreportspec
@@ -841,7 +909,7 @@ class HasReportOptsNoUpdate a => HasReportOpts a where
     statuses = reportOpts.statusesNoUpdate
     {-# INLINE statuses #-}
 
-    depth :: ReportableLens' a (Maybe Int)
+    depth :: ReportableLens' a DepthSpec
     depth = reportOpts.depthNoUpdate
     {-# INLINE depth #-}
 
@@ -891,5 +959,5 @@ updateReportSpecWith = overEither reportOpts
 
 -- | Generate a ReportSpec from RawOpts and a provided day, or return an error
 -- string if there are regular expression errors.
-rawOptsToReportSpec :: Day -> RawOpts -> Either String ReportSpec
-rawOptsToReportSpec day = reportOptsToSpec day . rawOptsToReportOpts day
+rawOptsToReportSpec :: Day -> Bool -> RawOpts -> Either String ReportSpec
+rawOptsToReportSpec day coloronstdout = reportOptsToSpec day . rawOptsToReportOpts day coloronstdout
